#include <GSE.h>
#include <UnitTest++/UnitTest++.h>


TEST(Members){

  
  GSE_Configuration *pConf = new GSE_Configuration();
  pConf->SetWarningsEnabled(0);
  pConf->ReadConfig( "./Test.cfg");
  // valid bools
  CHECK_EQUAL(true,  pConf->GetBoolean("TestSection|bool1t"));
  CHECK_EQUAL(false, pConf->GetBoolean("TestSection|bool1f"));
  CHECK_EQUAL(true,  pConf->GetBoolean("TestSection|bool2t"));
  CHECK_EQUAL(false, pConf->GetBoolean("TestSection|bool2f"));
  CHECK_EQUAL(true,  pConf->GetBoolean("TestSection|bool3t"));
  CHECK_EQUAL(false, pConf->GetBoolean("TestSection|bool3f"));

  // invalid bools
  CHECK_EQUAL(false, pConf->GetBoolean("TestSection|boolInvalid"));
  CHECK_EQUAL(false, pConf->GetBoolean("TestSection|boolInvalid2"));
  CHECK_EQUAL(false, pConf->GetBoolean("TestSection|boolEmpty"));

  // valid floats
  CHECK_CLOSE(0.0f, pConf->GetFloat("TestSection|floatZero"),0.001f );
  CHECK_CLOSE(-0.33f, pConf->GetFloat("TestSection|floatNeg"),0.001f);
  CHECK_CLOSE(4.22f, pConf->GetFloat("TestSection|floatPos"),0.001f);

  // invalid floats
  CHECK_CLOSE(0.0f, pConf->GetFloat("TestSection|floatInvalid"),0.001f);
  CHECK_CLOSE(0.0f, pConf->GetFloat("TestSection|floatEmpty"),0.001f);


  // String tests
  CHECK_EQUAL( "1.03454", pConf->GetString("TestSection|stringNumber").c_str());
  CHECK_EQUAL( "well", pConf->GetString("TestSection|stringNormal").c_str());
  CHECK_EQUAL( "yeah babe", pConf->GetString("TestSection|stringWithSpace").c_str());
  CHECK_EQUAL( "yeah baby let's rock", pConf->GetString("TestSection|stringWithSpaceMulti").c_str());
  CHECK_EQUAL( "", pConf->GetString("TestSection|stringEmpty").c_str());
  
  //  Int checks
  CHECK_EQUAL( 0, pConf->GetInt("TestSection|intZero"));
  CHECK_EQUAL( -13, pConf->GetInt("TestSection|intNeg"));
  CHECK_EQUAL( 45, pConf->GetInt("TestSection|intPos"));
  CHECK_EQUAL( 0, pConf->GetInt("TestSection|intInvalid"));
  CHECK_EQUAL( 2147483647, pConf->GetInt("TestSection|intInvalid2"));
  CHECK_EQUAL( 0, pConf->GetInt("TestSection|intEmpty"));

  //  long checks
  CHECK_EQUAL( 0, pConf->GetLong("TestSection|longZero"));
  CHECK_EQUAL( -23423592839428342LL, pConf->GetLong("TestSection|longNeg"));
  CHECK_EQUAL( 23423592839428343LL, pConf->GetLong("TestSection|longPos"));
  CHECK_EQUAL( 0, pConf->GetLong("TestSection|longInvalid"));
  CHECK_EQUAL( -3453, pConf->GetLong("TestSection|longInvalid2"));
  CHECK_EQUAL( 0, pConf->GetLong("TestSection|longEmpty"));
  
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,0,0).m_pValues, pConf->GetVector<3>("TestSection|vector3Zero").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(3,4,5).m_pValues, pConf->GetVector<3>("TestSection|vector3Pos").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(-3.0,-2.0,-6.0).m_pValues, pConf->GetVector<3>("TestSection|vector3Neg").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(-3,5,0).m_pValues, pConf->GetVector<3>("TestSection|vector3Mixed").m_pValues, 3,0.001f);

  CHECK_ARRAY_CLOSE( GSE_Vector3(0,3,5).m_pValues, pConf->GetVector<3>("TestSection|vector3Invalid").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(3,0,5).m_pValues, pConf->GetVector<3>("TestSection|vector3Invalid2").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(3,5,0).m_pValues, pConf->GetVector<3>("TestSection|vector3Invalid3").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,5,0).m_pValues, pConf->GetVector<3>("TestSection|vector3Invalid4").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,0,0).m_pValues, pConf->GetVector<3>("TestSection|vector3Empty").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,0,0).m_pValues, pConf->GetVector<3>("TestSection|vector3Empty2").m_pValues, 3,0.001f);

  CHECK_ARRAY_CLOSE( GSE_Color(1.0,0.5,0.0,1.0).GetValues(), pConf->GetVector<4>("TestSection|vector4Normal").m_pValues, 4, 0.001f);
  delete pConf;
}

TEST(SectionNameMembers){
  
  GSE_Configuration *pConf = new GSE_Configuration();
  pConf->SetWarningsEnabled(0);
  pConf->SetSectionString("TestSectionName");
  pConf->SetSectionEndString("}");
  pConf->ReadConfig( "./Test.cfg");
  // valid bools
  CHECK_EQUAL(true,  pConf->GetBoolean("Wheebaby|bool1t"));
  CHECK_EQUAL(false, pConf->GetBoolean("Wheebaby|bool1f"));
  CHECK_EQUAL(true,  pConf->GetBoolean("Wheebaby|bool2t"));
  CHECK_EQUAL(false, pConf->GetBoolean("Wheebaby|bool2f"));
  CHECK_EQUAL(true,  pConf->GetBoolean("Wheebaby|bool3t"));
  CHECK_EQUAL(false, pConf->GetBoolean("Wheebaby|bool3f"));

  // invalid bools
  CHECK_EQUAL(false, pConf->GetBoolean("Wheebaby|boolInvalid"));
  CHECK_EQUAL(false, pConf->GetBoolean("Wheebaby|boolInvalid2"));
  CHECK_EQUAL(false, pConf->GetBoolean("Wheebaby|boolEmpty"));

  // valid floats
  CHECK_CLOSE(0.0f, pConf->GetFloat("Wheebaby|floatZero"),0.001f );
  CHECK_CLOSE(-0.33f, pConf->GetFloat("Wheebaby|floatNeg"),0.001f);
  CHECK_CLOSE(4.22f, pConf->GetFloat("Wheebaby|floatPos"),0.001f);

  // invalid floats
  CHECK_CLOSE(0.0f, pConf->GetFloat("Wheebaby|floatInvalid"),0.001f);
  CHECK_CLOSE(0.0f, pConf->GetFloat("Wheebaby|floatEmpty"),0.001f);


  // String tests
  CHECK_EQUAL( "1.03454", pConf->GetString("Wheebaby|stringNumber").c_str());
  CHECK_EQUAL( "well", pConf->GetString("Wheebaby|stringNormal").c_str());
  CHECK_EQUAL( "yeah babe", pConf->GetString("Wheebaby|stringWithSpace").c_str());
  CHECK_EQUAL( "yeah baby let's rock", pConf->GetString("Wheebaby|stringWithSpaceMulti").c_str());
  CHECK_EQUAL( "", pConf->GetString("Wheebaby|stringEmpty").c_str());
  
  //  Int checks
  CHECK_EQUAL( 0, pConf->GetInt("Wheebaby|intZero"));
  CHECK_EQUAL( -13, pConf->GetInt("Wheebaby|intNeg"));
  CHECK_EQUAL( 45, pConf->GetInt("Wheebaby|intPos"));
  CHECK_EQUAL( 0, pConf->GetInt("Wheebaby|intInvalid"));
  CHECK_EQUAL( 2147483647, pConf->GetInt("Wheebaby|intInvalid2"));
  CHECK_EQUAL( 0, pConf->GetInt("Wheebaby|intEmpty"));

  //  long checks
  CHECK_EQUAL( 0, pConf->GetLong("Wheebaby|longZero"));
  CHECK_EQUAL( -23423592839428342LL, pConf->GetLong("Wheebaby|longNeg"));
  CHECK_EQUAL( 23423592839428343LL, pConf->GetLong("Wheebaby|longPos"));
  CHECK_EQUAL( 0, pConf->GetLong("Wheebaby|longInvalid"));
  CHECK_EQUAL( -3453, pConf->GetLong("Wheebaby|longInvalid2"));
  CHECK_EQUAL( 0, pConf->GetLong("Wheebaby|longEmpty"));
  
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,0,0).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Zero").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(3,4,5).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Pos").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(-3.0,-2.0,-6.0).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Neg").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(-3,5,0).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Mixed").m_pValues, 3,0.001f);

  CHECK_ARRAY_CLOSE( GSE_Vector3(0,3,5).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Invalid").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(3,0,5).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Invalid2").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(3,5,0).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Invalid3").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,5,0).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Invalid4").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,0,0).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Empty").m_pValues, 3,0.001f);
  CHECK_ARRAY_CLOSE( GSE_Vector3(0,0,0).m_pValues, pConf->GetVector<3>("Wheebaby|vector3Empty2").m_pValues, 3,0.001f);

  CHECK_ARRAY_CLOSE( GSE_Color(1.0,0.5,0.0,1.0).GetValues(), pConf->GetVector<4>("Wheebaby|vector4Normal").m_pValues, 4, 0.001f);
  delete pConf;
}
TEST(EmptySection)
{
  
  GSE_Configuration *pConf = new GSE_Configuration();
  pConf->SetWarningsEnabled(0);
  pConf->SetSectionString("TestSectionName");
  pConf->SetSectionEndString("}");
  pConf->ReadConfig( "./Test.cfg");
  CHECK_EQUAL( false, pConf->GetBoolean("testi"));
  CHECK_CLOSE( 0.0f, pConf->GetFloat("testi"), 0.001f);
  CHECK_EQUAL( 0, pConf->GetInt("testi"));
  CHECK_EQUAL( 0, pConf->GetLong("testi"));
  CHECK_EQUAL( "", pConf->GetString("testi").c_str());
  CHECK_ARRAY_CLOSE( GSE_Vector3().m_pValues, pConf->GetVector<3>("testi").m_pValues, 3, 0.001f);
  delete pConf;
  
}
int main()
{
  return UnitTest::RunAllTests();

}
// ----------------------------------------------------------------- 
// $Log: GSE_configurationTest.cpp,v $
// Revision 1.2  2007/03/30 08:48:10  entity
// removed RootMembers test since they are not allowed anymore
//
// Revision 1.1  2007/03/09 10:44:07  entity
// initial version
//
//
// ----------------------------------------------------------------- 
