#include <GSE.h>
#include <UnitTest++/UnitTest++.h>
#include <iostream>
using std::cerr;
using std::endl;
/////////////////////////////////////////////////////////////////
TEST(RayIntersectsSphereTestTwoPoints)
{
  
  GSE_Vector3 vRayEnd(0,0,-7.0);
  GSE_Vector3 vRayStart(0,0,7.0);
  GSE_Sphere  sSphere(GSE_Vector3(0,0,0),2.0f);
  GSE_Vector3 vInt0, vInt1;
  unsigned int nNumHits = 0; 

  nNumHits = Geometry::RayIntersectsSphere( vRayStart, vRayEnd, &vInt0, 
					    &vInt1, sSphere);
  
  CHECK_EQUAL( (unsigned int)2, nNumHits );

  CHECK_CLOSE( 0.0f,  vInt0[GSE_Vector3::X], 0.001f );
  CHECK_CLOSE( 0.0f,  vInt0[GSE_Vector3::Y], 0.001f );
  CHECK_CLOSE( 2.0f,  vInt0[GSE_Vector3::Z], 0.001f );

  CHECK_CLOSE( 0.0f,  vInt1[GSE_Vector3::X], 0.001f );
  CHECK_CLOSE( 0.0f,  vInt1[GSE_Vector3::Y], 0.001f );
  CHECK_CLOSE( -2.0f, vInt1[GSE_Vector3::Z], 0.001f );
  
}
/////////////////////////////////////////////////////////////////
TEST(RayIntersectsSphereTestOnePoint)
{
  
  GSE_Vector3 vRayEnd(2.0,0,-7.0);
  GSE_Vector3 vRayStart(2.0,0,7.0);
  GSE_Sphere  sSphere(GSE_Vector3(0,0,0),2.0f);
  GSE_Vector3 vInt0, vInt1;
  unsigned int nNumHits = 0; 

  nNumHits = Geometry::RayIntersectsSphere( vRayStart, vRayEnd, &vInt0, 
					    &vInt1, sSphere);
  
  CHECK_EQUAL( (unsigned int)1, nNumHits);

  CHECK_CLOSE( 2.0f,  vInt0[GSE_Vector3::X], 0.001f );
  CHECK_CLOSE( 0.0f,  vInt0[GSE_Vector3::Y], 0.001f );
  CHECK_CLOSE( 0.0f,  vInt0[GSE_Vector3::Z], 0.001f );
  
  CHECK( vInt0 == vInt1 );
  
  
}
/////////////////////////////////////////////////////////////////
TEST(RayIntersectsSphereTestNoHits){
  
  GSE_Vector3 vRayEnd(3,0,-7.0);
  GSE_Vector3 vRayStart(3,0,7.0);
  GSE_Sphere  sSphere(GSE_Vector3(0,0,0),2.0f);
  GSE_Vector3 vInt0, vInt1;
  unsigned int nNumHits = 0; 

  nNumHits = Geometry::RayIntersectsSphere( vRayStart, vRayEnd, &vInt0, 
					    &vInt1, sSphere);
  
  CHECK_EQUAL( (unsigned int)0, nNumHits );
  
}
/////////////////////////////////////////////////////////////////
TEST(MoreRayHitsSphereTests)
{
#define RADIUS 2.0f
  
  GSE_Sphere sSphere(GSE_Vector3(),RADIUS);
  GSE_Vector3 vRayStart; // changes a bit
  GSE_Vector3 vRayEnd(0,0,0); // Always origo
  GSE_Vector3 vInt0, vInt1;

  vRayStart[GSE_Vector3::X] = 10.5f;
  vRayStart[GSE_Vector3::Y] = 10.5f;
  vRayStart[GSE_Vector3::Z] = 10.5f;
  
  GSE_Vector3 vHit = (vRayStart - vRayEnd).ToUnit()*RADIUS;
  Geometry::RayIntersectsSphere( vRayStart, vRayEnd, &vInt0, 
				 &vInt1, sSphere);

  CHECK_CLOSE( vHit[GSE_Vector3::X],  vInt0[GSE_Vector3::X], 0.001f );
  CHECK_CLOSE( vHit[GSE_Vector3::Y],  vInt0[GSE_Vector3::Y], 0.001f );
  CHECK_CLOSE( vHit[GSE_Vector3::Z],  vInt0[GSE_Vector3::Z], 0.001f );
  
  
  vRayStart[GSE_Vector3::X] = 10.5f;
  vRayStart[GSE_Vector3::Y] = 20.5f;
  vRayStart[GSE_Vector3::Z] = 10.5f;
  
  vHit = (vRayStart - vRayEnd).ToUnit()*RADIUS;
  Geometry::RayIntersectsSphere( vRayStart, vRayEnd, &vInt0, 
				 &vInt1, sSphere);

  CHECK_CLOSE( vHit[GSE_Vector3::X],  vInt0[GSE_Vector3::X], 0.001f );
  CHECK_CLOSE( vHit[GSE_Vector3::Y],  vInt0[GSE_Vector3::Y], 0.001f );
  CHECK_CLOSE( vHit[GSE_Vector3::Z],  vInt0[GSE_Vector3::Z], 0.001f );
}
/////////////////////////////////////////////////////////////////
TEST(LineIntersectsAABB)
{
  GSE_Plane plane;
  GSE_Vector3 vStart,vEnd;
  GSE_Vector3 vPoint(0,0,0);
  GSE_Vector3 vNormal(1,0,0);
  plane.Calculate(vNormal,vPoint);
  ////////////////////
  vStart = GSE_Vector3(-1.01,0,0);
  vEnd = GSE_Vector3(1.01,0,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
  ////////////////////
  vStart = GSE_Vector3(0.001,0,0);
  vEnd = GSE_Vector3(1.01,0,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::FRONT );
  ////////////////////
  vStart = GSE_Vector3(1.01,0,0);
  vEnd = GSE_Vector3(2.01,0,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::FRONT );
  ////////////////////
  vStart = GSE_Vector3(-0.001,0,0);
  vEnd = GSE_Vector3(-1.01,0,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::BEHIND );
  ////////////////////
  vStart = GSE_Vector3(-1.01,0,0);
  vEnd = GSE_Vector3(-2.01,0,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::BEHIND );
  ////////////////////
  vStart = GSE_Vector3(0,-1.03,0);
  vEnd = GSE_Vector3(0,2.5,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
  ////////////////////
  vStart = GSE_Vector3(0,-1.03,0);
  vEnd = GSE_Vector3(0,-2.5,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
  ////////////////////
  vStart = GSE_Vector3(0,1.03,0);
  vEnd = GSE_Vector3(0,2.5,0);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
  ////////////////////
  vStart = GSE_Vector3(0,0,-1.03);
  vEnd = GSE_Vector3(0,0,-2.5);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
  ////////////////////
  vStart = GSE_Vector3(0,0,1.03);
  vEnd = GSE_Vector3(0,0,-2.5);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
  ////////////////////
  vStart = GSE_Vector3(0,0,-1.03);
  vEnd = GSE_Vector3(0,0,2.5);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
  ////////////////////
  vStart = GSE_Vector3(0,0,1.03);
  vEnd = GSE_Vector3(0,0,2.5);
  CHECK( Geometry::LineIntersectsPlane( plane, vStart, vEnd ) == Geometry::INTERSECTS );
}
/////////////////////////////////////////////////////////////////
TEST(PlaneIntersectsAABB_X_Axis)
{
  GSE_Plane plane;
  GSE_AxisAlignedBox aaBox;
  GSE_Vector3 vPoint;
  GSE_Vector3 vNormal;
  ////////////////////
  aaBox.SetPosition(0,0,0);
  aaBox.SetWidth( 1.0f);
  aaBox.SetHeight( 1.0f);
  aaBox.SetLength( 1.0f);
  ////////////////////
  /// The X-axis , plane normal towards negative x
  vNormal = GSE_Vector3(-1,0,0);
  vPoint  = GSE_Vector3(0.65f,0.0,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  vPoint  = GSE_Vector3(0.501f,0.0,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  vPoint  = GSE_Vector3(0.50f,0.0,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.499f,0.0,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );
  
  vPoint  = GSE_Vector3(0.45f,0.0,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(-0.45f,0.0,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(-0.499f,0.0,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(-0.50f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(-0.501f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );  

  vPoint  = GSE_Vector3(-0.65f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );  
  ////////////////////
  /// The X-axis , plane normal towards positive x
  vNormal = GSE_Vector3(1,0,0);
  vPoint  = GSE_Vector3(0.65f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );

  vPoint  = GSE_Vector3(0.501f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );

  vPoint  = GSE_Vector3(0.50f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.499f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );
  
  vPoint  = GSE_Vector3(0.45f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(-0.499f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(-0.50f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(-0.501f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );  

  vPoint  = GSE_Vector3(-0.65f,0.65,0.65);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );  

  
}
/////////////////////////////////////////////////////////////////
TEST(PlaneIntersectsAABB_Y_Axis)
{
  GSE_Plane plane;
  GSE_AxisAlignedBox aaBox;
  GSE_Vector3 vPoint;
  GSE_Vector3 vNormal;
  ////////////////////
  aaBox.SetPosition(0,0,0);
  aaBox.SetWidth( 1.0f);
  aaBox.SetHeight( 1.0f);
  aaBox.SetLength( 1.0f);
  ////////////////////
  /// The Y-axis , plane normal towards negative y
  vNormal = GSE_Vector3(0,-1,0);
  vPoint  = GSE_Vector3(0.0, 0.65f, 0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  vPoint  = GSE_Vector3(0.0f,0.501f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  vPoint  = GSE_Vector3(0.0f,0.50f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,0.499f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );
  
  vPoint  = GSE_Vector3(0.0f,0.45f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,-0.45f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0,-0.499f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0f,-0.50f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0, -0.501f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );  

  vPoint  = GSE_Vector3(0.0,-0.65f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );  
  ////////////////////
  /// The Y-axis , plane normal towards positive y
  vNormal = GSE_Vector3(0,1,0);
  vPoint  = GSE_Vector3(0.0, 0.65f, 0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );

  vPoint  = GSE_Vector3(0.0f,0.501f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );

  vPoint  = GSE_Vector3(0.0f,0.50f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,0.499f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );
  
  vPoint  = GSE_Vector3(0.0f,0.45f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,-0.45f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0,-0.499f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0f,-0.50f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0, -0.501f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );  

  vPoint  = GSE_Vector3(0.0,-0.65f,0.0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );  
  
}
/////////////////////////////////////////////////////////////////
TEST(PlaneIntersectsAABB_Z_Axis)
{
  GSE_Plane plane;
  GSE_AxisAlignedBox aaBox;
  GSE_Vector3 vPoint;
  GSE_Vector3 vNormal;
  ////////////////////
  aaBox.SetPosition(0,0,0);
  aaBox.SetWidth( 1.0f);
  aaBox.SetHeight( 1.0f);
  aaBox.SetLength( 1.0f);
  ////////////////////
  /// The Z-axis , plane normal towards negative z
  vNormal = GSE_Vector3(0,0,-1);
  vPoint  = GSE_Vector3(0.0, 0.0f, 0.65f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  vPoint  = GSE_Vector3(0.0f,0.0,0.501f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  vPoint  = GSE_Vector3(0.0f,0.0,0.50f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,0.0,0.499f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );
  
  vPoint  = GSE_Vector3(0.0f,0.0,0.45f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,0.0,-0.45f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0,0.0,-0.499f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0f,0.0,-0.50f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0,0.0,-0.501f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );  

  vPoint  = GSE_Vector3(0.0,0.0,-0.65f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );  
  ////////////////////
  /// The Z-axis , plane normal towards positive z
  vNormal = GSE_Vector3(0,0,1);
  vPoint  = GSE_Vector3(0.0,0.0,0.65f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );

  vPoint  = GSE_Vector3(0.0f,0.0,0.501f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );

  vPoint  = GSE_Vector3(0.0f,0.0,0.50f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,0.0,0.499f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );
  
  vPoint  = GSE_Vector3(0.0f,0.0,0.45f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0f,0.0,-0.45f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );

  vPoint  = GSE_Vector3(0.0,0.0,-0.499f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0f,0.0,-0.50f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS );  

  vPoint  = GSE_Vector3(0.0,0.0,-0.501f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );  

  vPoint  = GSE_Vector3(0.0,0.0,-0.65f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );  
  
}

/////////////////////////////////////////////////////////////////
TEST(PlaneIntersectsAABB_Arbitrary_Axis)
{
  GSE_Plane plane;
  GSE_AxisAlignedBox aaBox;
  GSE_Vector3 vPoint;
  GSE_Vector3 vNormal;
  ////////////////////
  aaBox.SetPosition(0,0,0);
  aaBox.SetWidth( 1.0f);
  aaBox.SetHeight( 1.0f);
  aaBox.SetLength( 1.0f);
  ////////////////////
  //    /  __   ^ y
  //   /\ |  |  |
  //  /   |__|  ---> x
  vNormal = GSE_Vector3(1,-1,0);
  vPoint  = GSE_Vector3(-0.65, 0.65f, 0.0f);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  ////////////////////
  //  \/ __   ^ y
  //   \|  |  |
  //    |\ |  ---> x
  //     
  vNormal = GSE_Vector3(1,1,0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS ); 

  
  ////////////////////
  //   \/  __   ^ y
  //   /  |  |  |
  //  /   |__|  ---> x
  vNormal = GSE_Vector3(-1,1,0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );
  
}
/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
TEST(PlaneIntersectsAABB_Arbitrary_Position)
{
  GSE_Plane plane;
  GSE_AxisAlignedBox aaBox;
  GSE_Vector3 vPoint;
  GSE_Vector3 vNormal;
  ////////////////////
  aaBox.SetPosition(30, 16.0,-42.3);
  aaBox.SetWidth( 1.0f);
  aaBox.SetHeight( 1.0f);
  aaBox.SetLength( 1.0f);
  ////////////////////
  //    /  __   ^ y
  //   /\ |  |  |
  //  /   |__|  ---> x
  vNormal = GSE_Vector3(1,-1,0);
  vPoint  = GSE_Vector3(-0.65, 0.65f, 0.0f) + aaBox.GetPosition();
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::FRONT );

  ////////////////////
  //  \/ __   ^ y
  //   \|  |  |
  //    |\ |  ---> x
  //     
  vNormal = GSE_Vector3(1,1,0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::INTERSECTS ); 

  
  ////////////////////
  //   \/  __   ^ y
  //   /  |  |  |
  //  /   |__|  ---> x
  vNormal = GSE_Vector3(-1,1,0);
  plane.Calculate(vNormal, vPoint);
  CHECK( Geometry::PlaneIntersectsAABB(plane, aaBox) == Geometry::BEHIND );
  
}

TEST(RayIntersectsOBB)
{
  GSE_OrientedBox obBox;
  GSE_Vector3 vUp(0,1,0);
  GSE_Vector3 vForward(0,0,-1);
  GSE_Vector3 vRight(1,0,0);
  GSE_Vector3 vRayStart(0,0,0);
  GSE_Vector3 vRayDir(0,0,-1.0f);
  
  obBox.SetOrientation( vUp, vForward, vRight);
  obBox.SetWidth(2.0f);
  obBox.SetHeight(2.0f);
  obBox.SetLength(2.0f);
  obBox.SetPosition( 0,0, -3.0f);

  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);
  vRayDir = GSE_Vector3( -1,0,-1);
  vRayDir.ToUnit();
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);
  vRayStart[GSE_Vector3::X] = 0.5f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);
  vRayStart[GSE_Vector3::X] = 0.999f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);
  vRayStart[GSE_Vector3::X] = 1.0f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);
  vRayStart[GSE_Vector3::X] = 1.001f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);

  vRayStart[GSE_Vector3::X] = 3.0f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);

  vRayStart[GSE_Vector3::X] = 4.999f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);
  vRayStart[GSE_Vector3::X] = 5.0f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);
  vRayStart[GSE_Vector3::X] = 5.001f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);

  vRayStart[GSE_Vector3::X] = 7.0f;
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);

  vRayStart = GSE_Vector3(-1.0, 0, 0);
  vRayDir   = GSE_Vector3(0,0,-1);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);

  vRayStart = GSE_Vector3(-1.001, 0, 0);
  //vRayDir   = GSE_Vector3(0,0,-1);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);

  vRayStart = GSE_Vector3( 1.001, 0, 0);
  //vRayDir   = GSE_Vector3(0,0,-1);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);
  
  vRayStart = GSE_Vector3( -2.0f, 0, -3.0f);
  vRayDir   = GSE_Vector3( 1,0,0);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);

  vRayStart = GSE_Vector3( -2.0f, 0, -4.0f);
  //vRayDir   = GSE_Vector3( 1,0,0);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);

  vRayStart = GSE_Vector3( -2.0f, 0, -4.001f);
  //vRayDir   = GSE_Vector3( 1,0,0);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);

  vRayStart = GSE_Vector3( -2.0f, 0, -1.999f);
  //vRayDir   = GSE_Vector3( 1,0,0);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 0);

  vRayStart = GSE_Vector3( 0, 3.0f, -3.0f);
  vRayDir   = GSE_Vector3(0,-1,0);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);

  vRayDir   = GSE_Vector3(0,-1,0);
  vRayStart = GSE_Vector3( 0, 0.0f, -3.0f);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);

  vRayStart = GSE_Vector3( -0.99, 3.0f, -3.0f);
  CHECK( Geometry::RayIntersectsOBB(vRayStart, vRayDir, obBox, NULL) == 1);
  
}
TEST( CalculateOBB )
{
  float pVertices[] = { 
    -1,-2, 1,
     1, 0, 2,
     2,-1, 3,
     2,-1, 2
  };
  
  GSE_Vector3 vR(-0.833f, -0.330f,  -0.443f);
  GSE_Vector3 vS(-0.257f,  0.941f,  -0.218f);
  GSE_Vector3 vT( 0.489f, -0.0675f, -0.870f);

  GSE_OrientedBox obBox = Geometry::CalculateOrientedBoundingBox( pVertices,4);
  
  CHECK_CLOSE( 1.0f, fabs(vR.Dot(obBox.GetForwardVector())), 0.001f);
  CHECK_CLOSE( 1.0f, fabs(vS.Dot(obBox.GetRightVector())), 0.001f);
  CHECK_CLOSE( 1.0f, fabs(vT.Dot(obBox.GetUpVector())), 0.001f);
  
}
TEST( CalculateOBB2 )
{
  float pVertices[] = { 
    0,0,-100,
    0,0,-50,
    0,0,50,
    0,0,100,
    0,10,0,
    0,5,0,
    0,-10,0,
    5,0,0,
    -5,0,0
  };

  GSE_OrientedBox obBox = Geometry::CalculateOrientedBoundingBox( pVertices,9);
  cerr << obBox.GetForwardVector() << endl;

  CHECK_CLOSE( 200.0f, fabs(obBox.GetLength()), 0.001f);
  CHECK_CLOSE( 20.0f, fabs(obBox.GetWidth()), 0.001f);
  CHECK_CLOSE( 10.0f, fabs(obBox.GetHeight()), 0.001f);
  
}
TEST( MergeOBBs )
{
  GSE_OrientedBox obBoxA;
  GSE_OrientedBox obBoxB;
  GSE_OrientedBox obBoxC;

  obBoxA.SetWidth( 1.0 );
  obBoxA.SetHeight( 1.0 );
  obBoxA.SetLength( 1.0 );  
  obBoxA.SetPosition(-1.0, 0.0, -1.0);
  obBoxA.CalculatePlanes();
  obBoxA.CalculateCorners();
  obBoxB.SetWidth( 1.0 );
  obBoxB.SetHeight( 1.0 );
  obBoxB.SetLength( 1.0 );  
  obBoxB.SetPosition(1.0, 0.0, 1.0);  
  obBoxB.CalculatePlanes();
  obBoxB.CalculateCorners();
  cerr << "starting merge test" << endl;
  obBoxC = Geometry::MergeOrientedBoxes( obBoxA, obBoxB );
  cerr << obBoxC << endl;
  
}
int main()
{
  return UnitTest::RunAllTests();

}
/////////////////////////////////////////////////////////////////
// $Log: GSE_geometryTest.cpp,v $
// Revision 1.10  2007/05/20 15:11:29  entity
// mergeOBB added
//
// Revision 1.9  2007/05/11 11:47:19  entity
// new tests
//
// Revision 1.8  2007/05/11 06:59:41  entity
// new tests
//
// Revision 1.7  2007/05/10 21:01:53  entity
// new tests
//
// Revision 1.6  2007/03/30 08:41:30  entity
// added LineIntersectsAABB test
//
// Revision 1.5  2007/03/30 08:13:32  entity
// more tests added
//
// Revision 1.4  2007/03/30 08:03:00  entity
// plane-AABB intersection tests added
//
// Revision 1.3  2007/03/26 18:24:22  entity
// more tests added
//
// Revision 1.2  2007/02/27 12:00:32  entity
// isolated tests for 2 hits, 1 hit and 0 hits + corrected the macro params
//
// Revision 1.1  2007/02/27 11:06:27  entity
// first UnitTest++ test
//
/////////////////////////////////////////////////////////////////
