#include <GSE.h>
#include <UnitTest++/UnitTest++.h>
#include <iostream>
using std::cerr;
using std::endl;
/////////////////////////////////////////////////////////////////

TEST( PrincipalComponentAnalysis )
{
  /////////////////////////////////////////////////////////////////
  // Data and results
  float pVertices[] = { 
    -1,-2, 1,
     1, 0, 2,
     2,-1, 3,
     2,-1, 2
  };
  GSE_Matrix3x3f mCovarActual;
   mCovarActual(0,0) = 1.5f;
   mCovarActual(0,1) = 0.5f;
   mCovarActual(0,2) = 0.75f;
   mCovarActual(1,0) = 0.5f;
   mCovarActual(1,1) = 0.5f;
   mCovarActual(1,2) = 0.25f;
   mCovarActual(2,0) = 0.75f;
   mCovarActual(2,1) = 0.25f;
   mCovarActual(2,2) = 0.5f;
   

  GSE_Matrix3x3f mEigenActual;

  GSE_Vector3 vR(-0.833f, -0.330f,  -0.443f);
  GSE_Vector3 vS(-0.257f,  0.941f,  -0.218f);
  GSE_Vector3 vT( 0.489f, -0.0675f, -0.870f);
   
  
  float fLambda1Actual = 2.097f;
  float fLambda2Actual = 0.3055f;
  float fLambda3Actual = 0.09756f;  
  
  /////////////////////////////////////////////////////////////////
  // Actual test
  GSE_Matrix3x3f mCovar = Math::CovarianceMatrix( pVertices, 4  );
  CHECK_ARRAY_CLOSE( mCovarActual.GetArray(),  mCovar.GetArray(), 9, 0.001f);
  
  GSE_Matrix3x3f mEigenTest;
  float fLambda1,fLambda2,fLambda3;  
  Math::CalculateEigensystem( mCovarActual, 
			      fLambda1, fLambda2,fLambda3,
			      mEigenTest);
  
  // Since vectors can be parallel, but point in opposite directions,
  // we test their dot products.

  CHECK_CLOSE( 1.0f,fabs(Math::GetColumnVector(mEigenTest,0).Dot(vR)), 0.001f);
  CHECK_CLOSE( 1.0f,fabs(Math::GetColumnVector(mEigenTest,1).Dot(vS)), 0.001f);
  CHECK_CLOSE( 1.0f,fabs(Math::GetColumnVector(mEigenTest,2).Dot(vT)), 0.001f);
  //CHECK_ARRAY_CLOSE( mEigenActual.GetArray(),  mEigenTest.GetArray(), 9, 0.001f);
  
  
  
  CHECK_CLOSE( fLambda1Actual, fLambda1, 0.001f);
  CHECK_CLOSE( fLambda2Actual, fLambda2, 0.001f);
  CHECK_CLOSE( fLambda3Actual, fLambda3, 0.001f);

  
}
int main()
{
  return UnitTest::RunAllTests();

}
/////////////////////////////////////////////////////////////////
// $Log: GSE_mathTest.cpp,v $
// Revision 1.4  2007/05/11 12:00:26  entity
// eigensystem test
//
// Revision 1.3  2007/05/11 08:18:28  entity
// a working test
//
// Revision 1.2  2007/05/11 07:07:37  entity
// compiles
//
// Revision 1.1  2007/05/11 07:05:51  entity
// math test
//
/////////////////////////////////////////////////////////////////
