#include <GSE.h>
#include <UnitTest++/UnitTest++.h>

float g_afZero[16] = { 0,0,0,0,
		      0,0,0,0,
		      0,0,0,0,
		      0,0,0,0 };

float g_afIdentity[16] = { 1,0,0,0,
			  0,1,0,0,
			  0,0,1,0,
			  0,0,0,1 };
float g_afTransposeTestA[16] = { 11,12,13,14,
			     21,22,23,24,
			     31,32,33,34,
			     41,42,43,44 };

float g_afTransposeTestB[16] = { 11,21,31,41,
			     12,22,32,42,
			     13,23,33,43,
			     14,24,34,44 };

#define NUM_ITEMS 16
#define PRECISION 0.0001f
TEST(MatrixTest){
  
  GSE_Matrix4x4f mMatrix;
  // Zero matrix
  CHECK_ARRAY_CLOSE(  g_afZero,  mMatrix.GetArray(), NUM_ITEMS, PRECISION);
  // Identity matrix
  mMatrix = GSE_Matrix4x4f::Identity();
  CHECK_ARRAY_CLOSE(  g_afIdentity,  mMatrix.GetArray(), NUM_ITEMS, PRECISION);
  
  // 
  mMatrix = GSE_Matrix4x4f::Identity();
  CHECK_ARRAY_CLOSE(  g_afIdentity,  mMatrix.GetArray(), NUM_ITEMS, PRECISION);
  
  GSE_Matrix4x4f mOrig(g_afTransposeTestA);
  GSE_Matrix4x4f mTransposed = mOrig.Transposed();

  CHECK_ARRAY_CLOSE(  g_afTransposeTestA,  mOrig.GetArray(),      NUM_ITEMS, PRECISION);
  CHECK_ARRAY_CLOSE(  g_afTransposeTestB,  mTransposed.GetArray(), NUM_ITEMS, PRECISION);
  
}
TEST( TranslationMatrix )
{
  GSE_Matrix4x4f mTranslation = Math::TranslationMatrix( 1,2,3 );
  CHECK_CLOSE( 1.0, mTranslation.At(0,3), PRECISION );
  CHECK_CLOSE( 2.0, mTranslation.At(1,3), PRECISION );
  CHECK_CLOSE( 3.0, mTranslation.At(2,3), PRECISION );
}
TEST( RotationMatrix )
{
#define RADIANS DEG2RAD(45.0f)

  GSE_Matrix4x4f mRotationX;
  GSE_Matrix4x4f mRotationY;
  GSE_Matrix4x4f mRotationZ;
  GSE_Matrix4x4f mRotationCombined;
  GSE_Matrix4x4f mRotationXYZ;
  float afRotationX[16] = { 1, 0,	      0,            0,
			    0, cos(RADIANS), -sin(RADIANS), 0,
			    0, sin(RADIANS),  cos(RADIANS), 0,
			    0, 0,             0,            1};

  float afRotationY[16] = { cos(RADIANS), 0,sin(RADIANS),0,
			    0,            1, 0,           0,
			    -sin(RADIANS), 0, cos(RADIANS),0,
			    0,            0, 0,           1};

  float afRotationZ[16] = { cos(RADIANS),-sin(RADIANS),0,0,
			    sin(RADIANS),cos(RADIANS), 0,0,
			    0,           0,            1,0,
			    0,           0,            0,1};
  
  mRotationX= Math::RotationMatrix( RADIANS,0,0 );
// X -axis
//          |  1  0       0       0 |
//      M = |  0  cos(A) -sin(A)  0 |
//          |  0  sin(A)  cos(A)  0 |
//          |  0  0       0       1 |
// Y-axis
  mRotationY= Math::RotationMatrix( 0,RADIANS,0 );
//          |  cos(A)  0  sin(A)  0 |
//      M = |  0       1   0       0 |
//          |  -sin(A)  0  cos(A)  0 |
//          |  0       0   0       1 |
  mRotationZ= Math::RotationMatrix( 0,0,RADIANS);
// Z-axis
//         |  cos(A)  -sin(A)   0   0 |
//     M = |  sin(A)   cos(A)   0   0 |
//         |  0        0        1   0 |
//         |  0        0        0   1 |


  mRotationCombined = mRotationZ * mRotationY * mRotationX;
  mRotationXYZ = Math::RotationMatrix( RADIANS, RADIANS, RADIANS );

  CHECK_ARRAY_CLOSE( afRotationX,  mRotationX.GetArray(),      NUM_ITEMS, PRECISION);
  CHECK_ARRAY_CLOSE( afRotationY,  mRotationY.GetArray(),      NUM_ITEMS, PRECISION);
  CHECK_ARRAY_CLOSE( afRotationZ,  mRotationZ.GetArray(),      NUM_ITEMS, PRECISION);
  CHECK_ARRAY_CLOSE( mRotationCombined.GetArray(),  mRotationXYZ.GetArray(), NUM_ITEMS, PRECISION);
				       
  
}
TEST( QuaterinonToMatrixAndBackAgain )
{
  GSE_Quaternion qStart;
  GSE_Quaternion qBack;
  GSE_Matrix4x4f mRotation;
  GSE_Matrix4x4f mRotation2;
  
  qStart[GSE_Quaternion::X] =
  qStart[GSE_Quaternion::Y] =
  qStart[GSE_Quaternion::Z] = 0.0f;
  qStart[GSE_Quaternion::W] = 1.0f;
  mRotation = qStart.ToMatrix();
  qBack = Math::RotationMatrixToQuaternion( mRotation );
  
  CHECK_CLOSE( qStart[GSE_Quaternion::X], qBack[GSE_Quaternion::X], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::Y], qBack[GSE_Quaternion::Y], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::Z], qBack[GSE_Quaternion::Z], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::W], qBack[GSE_Quaternion::W], PRECISION);
  qStart.CreateFromAxisAngle( 0,1,0,180.0f);
  mRotation = qStart.ToMatrix();
  qBack = Math::RotationMatrixToQuaternion( mRotation );
  
  mRotation2 = RotationMatrix(0,DEG2RAD(180.0f),0);
  CHECK_ARRAY_CLOSE(  mRotation2.GetArray(),  mRotation2.GetArray(),      NUM_ITEMS, PRECISION);
  
  CHECK_CLOSE( qStart[GSE_Quaternion::X], qBack[GSE_Quaternion::X], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::Y], qBack[GSE_Quaternion::Y], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::Z], qBack[GSE_Quaternion::Z], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::W], qBack[GSE_Quaternion::W], PRECISION);

  mRotation = Math::RotationMatrix( 0, DEG2RAD(90.0),0);
  qStart.CreateFromAxisAngle( 0,1.0,0,90.0f);
  qBack = Math::RotationMatrixToQuaternion( mRotation );
  
  CHECK_CLOSE( qStart[GSE_Quaternion::X], qBack[GSE_Quaternion::X], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::Y], qBack[GSE_Quaternion::Y], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::Z], qBack[GSE_Quaternion::Z], PRECISION);
  CHECK_CLOSE( qStart[GSE_Quaternion::W], qBack[GSE_Quaternion::W], PRECISION);
}
TEST( QuaternionInverse )
{
  GSE_Quaternion qOne, qTwo, qThree;
  qOne.CreateFromAxisAngle( 0,1,0,DEG2RAD(38.12f) );
  qTwo.CreateFromAxisAngle( 0,1,0,DEG2RAD(-38.12f) );
  qThree = qTwo.Inverse();
  
  CHECK_CLOSE( qOne[GSE_Quaternion::X], qThree[GSE_Quaternion::X], PRECISION);
  CHECK_CLOSE( qOne[GSE_Quaternion::Y], qThree[GSE_Quaternion::Y], PRECISION);
  CHECK_CLOSE( qOne[GSE_Quaternion::Z], qThree[GSE_Quaternion::Z], PRECISION);
  CHECK_CLOSE( qOne[GSE_Quaternion::W], qThree[GSE_Quaternion::W], PRECISION);

}
TEST( MatrixVectorMult )
{
  float fArray[] = { 0,0,0,1.0f};
  float fArrayResult[] = { 0,0,0,1.0f};
  float fArrayResult2[] = { -3,0,0,1.0f};
  GSE_Vector<float,4> vVec;
  // Assign vector values
  for(int i=0;i<4;i++)    
    vVec[i] = fArray[i];

  GSE_Matrix4x4f mMatrix = GSE_Matrix4x4f::Identity();
  CHECK_ARRAY_CLOSE(  fArrayResult,  vVec.m_pValues, 4, 0.01f);
  vVec = mMatrix * vVec ;
  CHECK_ARRAY_CLOSE(  fArrayResult,  vVec.m_pValues, 4, 0.01f);
  mMatrix(0,3) = -3.0f;
  vVec = mMatrix * vVec ;
  CHECK_ARRAY_CLOSE(  fArrayResult2,  vVec.m_pValues, 4, 0.01f);
  
}
int main()
{
  return UnitTest::RunAllTests();

}
// ----------------------------------------------------------------- 
// $Log: GSE_quotmatrixTest.cpp,v $
// Revision 1.5  2007/05/09 11:27:51  entity
// matrix * vector test
//
// Revision 1.4  2007/03/15 16:29:39  entity
// Quaternion inverse added ( negative rotation)
//
// Revision 1.3  2007/03/15 13:52:06  entity
// more tests
//
// Revision 1.2  2007/03/15 11:44:11  entity
// Additional tests implemented
//
// Revision 1.1  2007/03/15 10:26:31  entity
// Initial import to cvs
//
// ----------------------------------------------------------------- 
